/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker.generators;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import fr.supelec.adlv.checker.DefaultAbstractCheckingCodeGenerator;
import fr.supelec.adlv.checker.GenerationException;
import fr.supelec.adlv.checker.Signal;
import fr.supelec.adlv.checker.Util;
import fr.supelec.adlv.checker.logic.Always;
import fr.supelec.adlv.checker.logic.And;
import fr.supelec.adlv.checker.logic.BackTo;
import fr.supelec.adlv.checker.logic.BinaryExpr;
import fr.supelec.adlv.checker.logic.Constant;
import fr.supelec.adlv.checker.logic.Expr;
import fr.supelec.adlv.checker.logic.Implies;
import fr.supelec.adlv.checker.logic.Interval;
import fr.supelec.adlv.checker.logic.Not;
import fr.supelec.adlv.checker.logic.Once;
import fr.supelec.adlv.checker.logic.Or;
import fr.supelec.adlv.checker.logic.Previous;
import fr.supelec.adlv.checker.logic.Since;
import fr.supelec.adlv.checker.logic.UnaryExpr;
import fr.supelec.adlv.checker.logic.Interval.SignalCombination;

/**
 * Lustre code generator from patterns. This class generates Lustre v4 code.
 *
 */
public class LustreGenerator extends DefaultAbstractCheckingCodeGenerator {
	private String latestExpr = null;
	private int internalSignalNumber = 1;
	private final PrintStream ps;
	private final String adlvFileName;
	private final String lustreFileName;
	private final List<String[]> neverClauseSignals = new ArrayList<String[]>();
	private final List<String[]> internalSignals = new ArrayList<String[]>();
	//private final Set<String> inputSignals = new HashSet<String>();

	private final String ctrlNode;
	private final String ctrlFile;
	private final List<String> ctrlNodeInputs;
	private final List<String> ctrlNodeOutputs;

	private final boolean debug;

	public LustreGenerator(
			String adlvFileName,
			List<String> inPorts,
			List<String> outPorts,
			String controllerFile,
			String controllerNode,
			File outFile,
			boolean debug) throws FileNotFoundException {
		ps = new PrintStream(new FileOutputStream(outFile));
		this.adlvFileName = adlvFileName;
		this.ctrlNodeInputs = inPorts;
		this.ctrlNodeOutputs = outPorts;
		this.ctrlNode = controllerNode;
		this.ctrlFile = controllerFile;
		this.lustreFileName = outFile.toString();
		this.debug = debug;
	}

	private String generateNewInternalSignal(String prefix) {
		return prefix + "_" + (internalSignalNumber++);
	}

	private String[] addInternalSignal(String sigName, String expr, String[] comments) {
		String[] sig = new String[comments.length +2];
		sig[0] = sigName; sig[1] = expr;
		for(int i=0; i<comments.length; i++)
			sig[i+2] = comments[i];
		internalSignals.add(sig);
		return sig;
	}

	public void generateNeverClause(Expr pattern, String[] comments) {
		pattern.accept(this);
		String[] sig = addInternalSignal(generateNewInternalSignal("never"),
				latestExpr, comments);
		neverClauseSignals.add(sig);
		//inputSignals.addAll(pattern.getInputSignals());
	}

	public void start() {
		ps.println("-- Lustre checker code for " + adlvFileName);
		ps.println("-- File " + lustreFileName);
		ps.println("-- Generated by MoDriVal checker generator");
		ps.println("-- $Id: LustreGenerator.java 217 2007-11-27 13:15:06Z jacquet_chr $ ");
		ps.println("--");
		ps.println("-- To compile, run:");
		ps.println("--        lustre check.lus check");
		ps.println("--");
		ps.println("-- Or to check, run:");
		ps.println("--        lesar check.lus check -forward -diag");
		ps.println();
		ps.println(
				"-- Past Temporal Logic Library\n" +
				"node previous(A: bool) returns (prev_A: bool);\n" +
				"let\n" +
				"\tprev_A = false -> pre(A);\n" +
				"tel\n\n" +

				"node always(A: bool) returns (always_A: bool);\n" +
				"let\n" +
				"\talways_A = if not A then false else (true -> pre(always_A));\n" +
				"tel\n\n" +

				"node since(B, A: bool) returns (B_since_A: bool);\n" +
				"let\n" +
				"\tB_since_A =\n" +
				"\t\tif A then B\n" +
				"\t\telse B and (false -> pre(B_since_A));\n" +
				"tel\n\n" +

				"node backto(B, A: bool) returns (B_backto_A: bool);\n" +
				"let\n" +
				"\tB_backto_A =\n" +
				"\t\tif A then B\n" +
				"\t\telse B and (true -> pre(B_backto_A));\n" +
				"tel\n\n" +

				"node once(A: bool) returns (once_A: bool);\n" +
				"let\n" +
				"\tonce_A = if A then true else (false -> pre(once_A));\n" +
				"tel\n\n" +

				"node implies(A, B: bool) returns (A_implies_B: bool);\n" +
				"let\n" +
				"\tA_implies_B = (not A) or B;\n" +
				"tel\n\n" +

				"node interval(ini, begin, end: bool) returns (intv: bool);\n" +
				"let\n" +
				"\tintv = begin or (ini -> pre(intv) and not end);\n" +
				"tel\n\n" +

				"node first(a, b, rearm: bool) returns (res: bool);\n" +
				"var armed: bool;\n" +
				"let\n" +
				"\tarmed = true -> (not res) and (rearm or pre(armed));\n" +
				"\tres = armed and (a or b);\n" +
				"tel\n\n" +

				"node last(a, b, rearm: bool) returns (res: bool);\n" +
				"var armed, seen_a, seen_b: bool;\n" +
				"let\n" +
				"\tarmed = true -> (not res) and (rearm or pre(armed));\n" +
				"\tseen_a = armed and (a or (false -> pre(seen_a)));\n" +
				"\tseen_b = armed and (b or (false -> pre(seen_b)));\n" +
				"\tres = armed and ((seen_a and b) or (seen_b and a));\n" +
				"tel\n" +

				"-- End of Past Temporal Logic Library\n\n"
				);
	}

	public void stop() throws GenerationException {
		// list of "never" signals
		boolean notFirstN = false;
		StringBuffer orExpr = new StringBuffer("ok = not(");
		List<String> sigDefs = new ArrayList<String>();
		List<String> intSigNames = new ArrayList<String>();

		for(String[] p : internalSignals) {
			intSigNames.add(p[0]);

			StringBuffer sigDef = new StringBuffer();
			for(int i=2; i<p.length; i++)
				sigDef.append("\t-- ").append(p[i]).append('\n');
			sigDef.append("\t" + p[0] + " = " + p[1] + ";");
			sigDefs.add(sigDef.toString());

			if(neverClauseSignals.contains(p)) {
				if(notFirstN) orExpr.append(" or ");
				else notFirstN = true;
				orExpr.append(p[0]);
			}
		}
		orExpr.append(");");

		// includes controller
		ps.println("-- Include controller file: " + ctrlFile);
		ps.println("-- Assume there is a node called \"" + ctrlNode + "\" inside, with prototype:");
		ps.println("--   node controller(" + Util.collectionToString(ctrlNodeInputs) + ": bool)");
		ps.println("--   returns (" + Util.collectionToString(ctrlNodeOutputs) + ": bool);");
		ps.println("-- Will fail if it is not the case.");
		ps.println("include \"" + ctrlFile + "\"");
		ps.println();

		// generate never and always code here in order to have only one
		// "failure = ..." line
		ps.println("-- Node for never/always clauses");
		ps.println(
				"node check(" + Util.collectionToString(ctrlNodeInputs) +
				(ctrlNodeInputs.size() > 0 ? ": bool" : "") +
				") returns (ok" +
				(debug ? ", " + Util.collectionToString(intSigNames) + ", " + Util.collectionToString(ctrlNodeOutputs) : "") +
				": bool);");

		if(!debug)
			ps.println("var " +
					Util.collectionToString(ctrlNodeOutputs) + ", " +
					Util.collectionToString(intSigNames) +
					": bool;");
		ps.println("let");

		ps.println("\t-- Invoke the controller");
		ps.println("\t(" + Util.collectionToString(ctrlNodeOutputs) + ") = ");
		ps.println("\t\t" + ctrlNode + "(" + Util.collectionToString(ctrlNodeInputs) + ");");
		ps.println();

		for(String sd : sigDefs) {
			ps.println(sd);
			ps.println();
		}

		ps.println("\t" + orExpr);
		ps.println("tel");

		ps.println("-- End of Lustre checking code");
		ps.close();

		if(ctrlNodeInputs.contains("ok") || ctrlNodeOutputs.contains("ok"))
			throw new GenerationException(this, "When generating Lustre code, "
					+ "\"ok\" must not be an I/O of the controller. Correct "
					+ "the ADLV file for the checker to compile.");
	}


	private static Util.StringConverter<Signal> signalToString =
		new Util.StringConverter<Signal>() {
			public String convert(Signal s) {
				String res = s.port.getPort().getName();
				if(!s.positive) res = "not(" + res + ")";
				return res;
			}

	};

	private String intervalBound(Set<Signal> signals, Interval.SignalCombination comb) {
		if(comb == SignalCombination.OR)
			return Util.collectionToString(signals, " or ", signalToString);

		if(signals.size() == 1)
			return signalToString.convert(Util.pickFromSingleton(signals));

		if(signals.size() >= 2) {
			String op;
			switch(comb) {
			case FIRST: op = "first"; break;
			case LAST: op = "last"; break;
			default: throw new RuntimeException("LustreGenerator.intervalBound(): unknown signal combination.");
			}

			Set<Signal> remaining = new HashSet<Signal>(signals);
			Signal sig = Util.pickFromSingleton(remaining);
			remaining.remove(sig);
			return op + "(" + signalToString.convert(sig) + ", " +
				intervalBound(remaining, comb) + ")";
		}

		throw new RuntimeException("LustreGenerator.intervalBound() failed.");
	}


	public void visit(Interval interval) {
		String begin = intervalBound(interval.getBegin(), interval.getBeginCombination());
		String end = intervalBound(interval.getEnd(), interval.getEndCombination());

		latestExpr =
			"interval(" + interval.getInitialActivationState() + ", " +
			begin + ", " + end + ")";

		if(debug) {
			String[] sig = addInternalSignal(generateNewInternalSignal("interval"),
					latestExpr, new String[] {"Interval: " + interval.toString()});
			latestExpr = sig[0];
		}
	}

	private void visit(BinaryExpr operation, String operator, boolean infix) {
		Expr patternA = operation.getA();
		patternA.accept(this);
		String exprA = latestExpr;

		Expr patternB = operation.getB();
		patternB.accept(this);
		String exprB = latestExpr;

		if(infix)
			latestExpr = "(" + exprA + " " + operator + " " + exprB + ")";
		else
			latestExpr = operator + "(" + exprA + ", " + exprB + ")";
	}

	private void visit(UnaryExpr operation, String func) {
		Expr patternA = operation.getExpr();
		patternA.accept(this);

		latestExpr = func + "(" + latestExpr + ")";
	}

	public void visit(And and) {
		visit(and, "and", true);
	}

	public void visit(Or or) {
		visit(or, "or", true);
	}

	@Override
	public void visit(Implies implies) {
		visit(implies, "implies", false);
	}

	public void visit(Not not) {
		visit(not, "not");
	}

	public void visit(Since since) {
		visit(since, "since", false);
	}

	@Override
	public void visit(BackTo backTo) {
		visit(backTo, "backto", false);
	}

	public void visit(Always always) {
		visit(always, "always");
	}


	public void visit(Previous previous) {
		visit(previous, "previous");
	}

	@Override
	public void visit(Once once) {
		visit(once, "once");
	}

	public void visit(Constant constant) {
		latestExpr = constant.getValue() ? "true" : "false";
	}

	public String getName() {
		return "Lustre Generator";
	}
}
