/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker.logic;

import java.util.HashSet;
import java.util.Set;

import fr.supelec.adlv.checker.PortInstance;
import fr.supelec.adlv.checker.Signal;
import fr.supelec.adlv.checker.Util;

/**
 * This class represents an interval during which a condition holds.
 *
 */
public class Interval extends Predicate {
	/**
	 * The signal upon which the interval begins.
	 */
	private final Set<Signal> begin;
	private final SignalCombination beginComb;


	/**
	 * The signal upon which the interval ends.
	 */
	private final Set<Signal> end;
	private final SignalCombination endComb;


	/**
	 * Indicates whether the system is inside the interval at startup.
	 */
	private final boolean initiallyActivated;

	public Interval(
			Set<Signal> begin, SignalCombination beginComb,
			Set<Signal> end, SignalCombination endComb,
			boolean initiallyActivated) {
		this.begin = begin;
		this.beginComb = beginComb;
		this.end = end;
		this.endComb = endComb;
		this.initiallyActivated = initiallyActivated;
	}

	/**
	 * Constructor for basic intervals that just have one begin signal and one
	 * end signal.
	 *
	 * @param beginP
	 * @param beginR
	 * @param endP
	 * @param endR
	 * @param ini
	 */
	public Interval(PortInstance beginP, boolean beginR, PortInstance endP, boolean endR, boolean ini) {
		begin = new HashSet<Signal>(1); // (initial) capacity : 1
		begin.add(new Signal(beginP, beginR));
		beginComb = SignalCombination.OR;
		end = new HashSet<Signal>(1);   // ditto
		end.add(new Signal(endP, endR));
		endComb = SignalCombination.OR;
		initiallyActivated = ini;
	}

	/**
	 * Returns the signal upon which the interval begins.
	 * @return the signal upon which the interval begins
	 */
	public Set<Signal> getBegin() {
		return begin;
	}

	public SignalCombination getBeginCombination() {
		return beginComb;
	}


	/**
	 * Returns the signal upon which the interval ends.
	 * @return the signal upon which the interval ends.
	 */
	public Set<Signal> getEnd() {
		return end;
	}

	public SignalCombination getEndCombination() {
		return endComb;
	}


	/**
	 * Returns the initial activation state, i.e. <code>true</code> if at
	 * startup, the system is <em>inside</em> the interval, <code>false</code>
	 * otherwise.
	 * @return the initial activation state
	 */
	public boolean getInitialActivationState() {
		return initiallyActivated;
	}

	@Override
	public void accept(ExprVisitor visitor) {
		visitor.visit(this);
	}


	@Override
	public Predicate neg() {
		return new Interval(end, endComb, begin, beginComb, !initiallyActivated);
	}

	@Override
	public boolean isNegationOf(Predicate p) {
		if(p instanceof Interval) {
			Interval ip = (Interval)p;
			return ip.begin.equals(end) && ip.end.equals(begin) &&
					ip.initiallyActivated != initiallyActivated;
		}
		return false;
	}

	@Override
	protected Expr develop(boolean cnf) {
		throw new RuntimeException("Should not call develop() on an interval!");
	}

	@Override
	public String toString() {
		return "[" + Util.collectionToString(begin) + ";; " +
				Util.collectionToString(end) + "[";
	}

	@Override
	public Set<PortInstance> getInputSignals() {
		Set<PortInstance> is = new HashSet<PortInstance>();
		for(Signal sig : begin)
			is.add(sig.port);
		for(Signal sig : end)
			is.add(sig.port);
		return is;
	}

	/*private Set<Signal> invertSet(Set<Signal> set) {
		Set<Signal> res = new HashSet<Signal>();
		for(Signal sig : set) {
			res.add(new Signal(sig.port, !sig.rising));
		}
	}*/

	public static enum SignalCombination {
		FIRST("fisrt"), LAST("last"), OR("unique");

		private final String name;

		private SignalCombination(String name) {
			this.name = name;
		}

		@Override
		public String toString() {
			return name;
		}
	}
}