/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;

public abstract class AdlvGenerator {

	public static final String STREAM_TYPE_INTERFACE_FILENAME = "StreamTypeInterface";
	public static final String STI_MODULE_NAME = "supelec";
	public static final String STI_MODULE_DIRECTORY = "Supelec";

	public static final String BOOLEAN_STREAM_TYPE = "BooleanStreamConsumer";
	public static final String DOUBLE_STREAM_TYPE = "DoubleStreamConsumer";
	public static final String LONG_STREAM_TYPE = "LongStreamConsumer";
	
	public static final String BOOLEAN_DUPLICATOR = "BooleanStreamDuplicator";
	public static final String DOUBLE_DUPLICATOR = "DoubleStreamDuplicator";
	public static final String LONG_DUPLICATOR = "LongStreamDuplicator";
	
	public static final String IMPL_SUFFIX = "Impl";
	public static final String MCCM_TARGET_SUFFIX = "_$(MCCM_TARGET)";
	
	public static final String CPP_EXTENSION = "cpp";
	public static final String HPP_EXTENSION = "h";
	public static final String IDL_EXTENSION = "idl";
	public static final String IDL3_EXTENSION = "idl3";
	public static final String MAKE_PROJECT_EXTENSION = "mpc";
	public static final String MAKE_WORKSPACE_EXTENSION = "mwc";
	
	public static final String EVENT_TYPE_DECLARATION_FILENAME = "EventTypeDeclaration";
	public static final String EVENT_TYPE_DECLARATION_DIRECTORY = "Events";
	public static final String EVENT_TYPE_FILE_SUFFIX = "Events";
	
	public static final String DYNAMIC_ROUTING_NAME = "DynamicRouting";
	public static final String ENVIRONMENT_NAME = "Environment";
	public static final String SIMULATION_NAME = "simulation";
	public static final String INSTALL_DIRECTORY = "../run";
	
	// Only one of them should be set to true.
	// Only STATIC_DEPLOYMENT is currently working
	public static final boolean ORBLESS_DEPLOYMENT = false;
	public static final boolean STATIC_DEPLOYMENT = true;
	public static final boolean CORBA_DEPLOYMENT = false;
	
	// Currently, static libraries are not working
	public static final boolean STATIC_LIBRARIES = false;
	public static final boolean SHARED_LIBRARIES = ! STATIC_LIBRARIES;

	/**
	 * Makes an association between the data type from the description
	 * file and the data type from the .idl3 generated files
	 * @param dataDescription the data type from the description file
	 * @return the data type used in the IDL3 generated files
	 */

	public static String getStreamType( String dataDescription ) {
		if( dataDescription.equals( "boolean" )) {
			return BOOLEAN_STREAM_TYPE;
		}
		else if( dataDescription.equals( "long" ) ) {
			return LONG_STREAM_TYPE;
		}
		else if( dataDescription.equals( "double" )) {
			return DOUBLE_STREAM_TYPE;
		}
		else {
			return null;
		}
	}

	public static String getDuplicatorName( String dataDescription ) {
		if( dataDescription.equals( "boolean" )) {
			return BOOLEAN_DUPLICATOR;
		}
		else if( dataDescription.equals( "long" ) ) {
			return LONG_DUPLICATOR;
		}
		else if( dataDescription.equals( "double" )) {
			return DOUBLE_DUPLICATOR;
		}
		else {
			return null;
		}
	}

	public static String getCorbaType( String dataDescription ) {
		if( dataDescription.equals( "boolean" )) {
			return "CORBA::Boolean";
		}
		else if( dataDescription.equals( "long" ) ) {
			return "CORBA::Long";
		}
		else if( dataDescription.equals( "double" )) {
			return "CORBA::Double";
		}
		else {
			return null;
		}
	}

	protected void openFile( String path, String filename, String extension ) throws IOException {
		closeFile();
		
		path_ = path;
		filename_ = filename;
		extension_ = extension;
		
		( new File( path_ ) ).mkdirs();
		out_ = new FileWriter( new File( path_ + "/" + filename_ + "." + extension_ ));
	}
	
	protected void closeFile() throws IOException {
		if( out_ != null ) {
			out_.close();
			out_ = null;
		}
	}
	
	protected void writeCppCopyrightComment() throws IOException {
		out_.write( "/*\n" );
		out_.write( " * Copyright Supelec 2007\n" );
		out_.write( " *\n" );
		out_.write( " * This file is automatically generated, do not edit\n" );
		out_.write( " * Generated on " + DateFormat.getDateTimeInstance().format( new Date() ) + "\n" );
		out_.write( " */\n" );
		out_.write( "\n" );
	}
	
	protected void writeXmlCopyrightComment() throws IOException {
		out_.write( "<!--\n" );
		out_.write( " * Copyright Supelec 2007\n" );
		out_.write( " *\n" );
		out_.write( " * This file is automatically generated, do not edit\n" );
		out_.write( " * Generated on " + DateFormat.getDateTimeInstance().format( new Date() ) + "\n" );
		out_.write( "-->\n" );
		out_.write( "\n" );
	}
	
	protected void writeMpcCopyrightComment() throws IOException {
		out_.write( "//\n" );
		out_.write( "// Copyright Supelec 2007\n" );
		out_.write( "//\n" );
		out_.write( "// This file is automatically generated, do not edit\n" );
		out_.write( "// Generated on " + DateFormat.getDateTimeInstance().format( new Date() ) + "\n" );
		out_.write( "//\n" );
		out_.write( "\n" );
	}
	
	protected void writeCppHeaderPrelude( String moduleName ) throws IOException {
		writeCppCopyrightComment();
		
		out_.write( "#ifndef _"
				  + moduleName.toUpperCase()
				  + "_"
				  + filename_.toUpperCase()
				  + "_"
				  + extension_.toUpperCase()
				  + "\n" );
		out_.write( "#define _"
				  + moduleName.toUpperCase()
				  + "_"
				  + filename_.toUpperCase()
				  + "_"
				  + extension_.toUpperCase()
				  + "\n" );
		out_.write( "\n" );
	}
	
	protected void writeCppHeaderPostlude() throws IOException {
		out_.write( "\n" );
		out_.write( "#endif\n" );
		out_.write( "\n" );
	}

	protected FileWriter out_ = null;
	protected String path_;
	protected String filename_;
	protected String extension_;
}
