/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.generator;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;

import fr.supelec.adlv.*;

import fr.supelec.adlv.parser.EmfAdlvHelper;

public class ApplicationComponentImplGenerator extends AdlvGenerator {
	ApplicationComponentImplGenerator( EmfAdlvHelper helper ) {
		helper_ = helper;
	}
	
	void generate( String resultsPath ) throws IOException {
		generateDynamicRoutingComponent( resultsPath );
		generateEnvironmentComponent( resultsPath );
	}
	
	private void generateDynamicRoutingComponent( String resultsPath ) throws IOException {
		Application application = helper_.getApplication();
		String applicationDirectory = resultsPath + "/" + DYNAMIC_ROUTING_NAME;
		( new File( applicationDirectory ) ).mkdir();
		
		String impl = DYNAMIC_ROUTING_NAME + "_impl";

		openFile( applicationDirectory, impl, HPP_EXTENSION );
		
		writeCppHeaderPrelude( application.getModule().getName() );
		
		out_.write( "#include \"microccm_" + DYNAMIC_ROUTING_NAME + "_executor." + HPP_EXTENSION + "\"\n" );
		out_.write( "#include \"microccm_" + DYNAMIC_ROUTING_NAME + "_CORBA." + HPP_EXTENSION + "\"\n" );
		out_.write( "\n" );
		
		out_.write( "class CCM_" + impl + "\n" );
		out_.write( "    : public CCM_" + DYNAMIC_ROUTING_NAME + "\n" );
		out_.write( "    , public CCM_" + DYNAMIC_ROUTING_NAME + "_SessionComponent_impl\n" );
		out_.write( "{\n" );
		out_.write( "public:\n" );
		out_.write( "    CCM_" + impl + "();\n" );
		out_.write( "\n" );
		out_.write( "    void configuration_complete()\n" );
		out_.write( "        throw( Components::InvalidConfiguration ) {}\n" );
		out_.write( "\n" );
		out_.write( "    void shutdown() {}\n" );
		out_.write( "\n" );

		// Find the initial connection
		ConnectionConfigurationStep initial = null;
		for( ConfigurationStep step : application.getInitialSteps() ) {
			if( step instanceof ComponentConfigurationStep ) continue;
			
			if( initial == null ) initial = ( ConnectionConfigurationStep ) step;
			else {
				System.err.println( "Fatal Error, initial configuration works currently only with one step" );
				System.exit( -1 );
			}
		}
		
		// Currently, we only support several inputs going to one output
		Port toPort = initial.getConnection().getToPort();

		
		ArrayList< Port > fromPorts = new ArrayList< Port >();
		for( Port p : application.getPorts() ) {
			// Looking for event consumers
			if( p instanceof ProducerPort ) continue;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) continue;
			
			// that are used for dynamic configuration
			if((( ConsumerPort ) p ).getSteps().isEmpty() ) continue;
			EventType e = (EventType) t;
			out_.write( "    void push_" + p.getName() + "( "
					  + e.getModule().getName() + "::"
					  + e.getName() + " * evt );\n" );
			out_.write( "\n" );
			
			for( ConfigurationStep step : (( ConsumerPort ) p ).getSteps() ) {
				if( step instanceof ComponentConfigurationStep ) continue;
				
				fromPorts.add((( ConnectionConfigurationStep ) step ).getConnection().getFromPort() );
				
				if( ! toPort.equals( (( ConnectionConfigurationStep ) step ).getConnection().getToPort() )) {
					System.err.println( "Fatal Error, dynamic configuration works currently only with one output port" );
					System.exit( -1 );
				}
			}
		}
		
		{
			TokenType t = toPort.getTokenType();
			if( ! ( t instanceof StreamType )) {
				System.err.println( "Fatal Error, dynamic configuration works currently only with stream ports" );
				System.exit( -1 );
			}
			StreamType s = ( StreamType ) t;
			out_.write( "private:\n" );
			//out_.write( "    " + STI_MODULE_NAME + "::CCM_"
			out_.write( "    " + "CCM_"
					  + getStreamType( s.getType().getName() ) + "_ptr current_input_;\n" );
			out_.write( "\n" );
		}
		
		for( Port p : fromPorts ) {
			boolean applicationPort = p.getApplication() != null;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				// StreamType
				StreamType s = (StreamType) t;
				if((   applicationPort && ( p instanceof ConsumerPort ))
				|| ( ! applicationPort && ( p instanceof ProducerPort ))) {
					out_.write( "public:\n" );
					out_.write( "    void new_" + p.getName() + "_value( "
							  + getCorbaType( s.getType().getName() ) +" value );\n" );
					//out_.write( "    " + STI_MODULE_NAME + "::CCM_"
					out_.write( "    " + "CCM_"
							  + getStreamType( s.getType().getName() ) + "_ptr "
							  + "get_" + p.getName() + "() {\n" );
					out_.write( "        return &" + p.getName() + "_provider_;\n" );
					out_.write( "    }\n" );
					out_.write( "\n" );
					out_.write( "private:\n" );
					out_.write( "    struct Provides_" + p.getName() + "\n" );
					//out_.write( "        : " + STI_MODULE_NAME + "::CCM_"
					out_.write( "        : " + "CCM_"
							  + getStreamType( s.getType().getName() ) + "\n" );
					out_.write( "    {\n" );
					out_.write( "        Provides_" + p.getName()
							  + "( CCM_" + impl + " * c )\n" );
					out_.write( "            : component_( c ) {}\n" );
					out_.write( "\n" );
					out_.write( "        void put( "
							  + getCorbaType( s.getType().getName() )
							  + " value );\n" );
					out_.write( "\n" );
					out_.write( "        CCM_" + impl + " * component_;\n" );
					out_.write( "    };\n" );
					out_.write( "\n" );
					out_.write( "    Provides_" + p.getName() + " "
							  + p.getName() + "_provider_;\n" );
					out_.write( "\n" );
				}
			}
		}
		
		out_.write( "\n" );
		out_.write( "};\n" );
		out_.write( "\n" );

		String home = DYNAMIC_ROUTING_NAME + "Home";
		String home_impl = home + "_impl";

		out_.write( "class " + home_impl + "\n" );
		out_.write( "    : public CCM_" + home + "\n" );
		out_.write( "{\n" );
		out_.write( "public:\n" );
		out_.write( "    Components::EnterpriseComponent_ptr create()\n" );
		out_.write( "        throw( Components::CCMException );\n" );
		out_.write( "};\n" );
		out_.write( "\n" );
		
		out_.write( "extern \"C\"\n" );
		out_.write( "{\n" );
		out_.write( "    Components::HomeExecutorBase * create_" + DYNAMIC_ROUTING_NAME + "_home();\n" );
		out_.write( "}\n" );
		out_.write( "\n" );
		
		writeCppHeaderPostlude();
		
		closeFile();
		
		openFile( applicationDirectory, impl, CPP_EXTENSION );
		
		writeCppCopyrightComment();
		
		out_.write( "#include \"" + impl + "." + HPP_EXTENSION + "\"\n" );
		out_.write( "\n" );

		// Constructor
		out_.write( "CCM_" + impl + "::CCM_" + impl + "()\n" );
		
		out_.write( "    : current_input_( get_" + initial.getConnection().getFromPort().getName() + "() )\n" );

		for( Port p : fromPorts ) {
			boolean applicationPort = p.getApplication() != null;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				// StreamType
				StreamType s = (StreamType) t;
				if((   applicationPort && ( p instanceof ConsumerPort ))
				|| ( ! applicationPort && ( p instanceof ProducerPort ))) {
					out_.write( "    , " + p.getName() + "_provider_( this )\n" );
				}
			}
		}
		out_.write( "{\n" );
		out_.write( "}\n" );
		out_.write( "\n" );
		
		for( Port p : application.getPorts() ) {
			// Looking for event consumers
			if( p instanceof ProducerPort ) continue;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) continue;
			
			// that are used for dynamic configuration
			if((( ConsumerPort ) p ).getSteps().isEmpty() ) continue;
			EventType e = (EventType) t;
			out_.write( "void CCM_" + impl + "::push_" + p.getName() + "( "
					  + e.getModule().getName() + "::"
					  + e.getName() + " * evt )\n" );
			out_.write( "{\n" );
			out_.write( "#ifdef TRACE_ADLV\n" );
			out_.write( "    std::clog << \"TRACE_ADLV "
					  + ": application : " + application.getName()
					  + ", consumes : " + p.getName()
					  + ", event received\\n\";\n" );
			out_.write( "#endif\n" );
			out_.write( "\n" );
			
			for( ConfigurationStep step : (( ConsumerPort ) p ).getSteps() ) {
				if( step instanceof ComponentConfigurationStep ) continue;
				
				out_.write( "    current_input_ = get_"
						  + (( ConnectionConfigurationStep ) step ).getConnection().getFromPort().getName()
						  + "();\n" );
			}
			out_.write( "}\n" );
			out_.write( "\n" );
		}

		for( Port p : fromPorts ) {
			boolean applicationPort = p.getApplication() != null;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				// StreamType
				StreamType s = (StreamType) t;
				if((   applicationPort && ( p instanceof ConsumerPort ))
				|| ( ! applicationPort && ( p instanceof ProducerPort ))) {
					out_.write( "void CCM_" + impl + "::new_" + p.getName() + "_value( "
							  + getCorbaType( s.getType().getName() ) +" value )\n" );
					out_.write( "{\n" );
					out_.write( "    if( current_input_ == get_" + p.getName() + "() ) {\n" );
					StreamType s2 = (StreamType) toPort.getTokenType();
					//out_.write( "        " + STI_MODULE_NAME + "::"
					out_.write( "        "
							  + getStreamType( s2.getType().getName() )
							  + "_ptr cl = component_context_->get_connection_"
							  + toPort.getName()
							  + "();\n" );
					
					out_.write( "        if( cl ) {\n" );
					out_.write( "#ifdef TRACE_ADLV\n" );
					out_.write( "            std::clog << \"TRACE_ADLV "
							  + ": application : " + application.getName()
							  + ", uses : " + toPort.getName()
							  + ", value sent : \" << value << \"\\n\";\n" );
					out_.write( "#endif\n" );
					out_.write( "\n" );

					out_.write( "            cl->put( value );\n" );
					out_.write( "        }\n" );
					out_.write( "    }\n" );
					out_.write( "}\n" );
					out_.write( "\n" );

					out_.write( "void CCM_" + impl + "::Provides_" + p.getName()
							  + "::put( "
							  + getCorbaType( s.getType().getName() )
							  + " value )\n" );
					out_.write( "{\n" );
					out_.write( "#ifdef TRACE_ADLV\n" );
					out_.write( "    std::clog << \"TRACE_ADLV "
							  + ": application : " + application.getName()
							  + ", sink : " + p.getName()
							  + ", value received : \" << value << \"\\n\";\n" );
					out_.write( "#endif\n" );
					out_.write( "\n" );
					out_.write( "    component_->new_" + p.getName() + "_value( value );\n" );
					out_.write( "}\n" );
					out_.write( "\n" );
				}
			}
		}
		
		out_.write( "//---------------------------------------------------------------------\n\n" );

		out_.write( "Components::EnterpriseComponent_ptr\n" );
		out_.write( home_impl + "::create() throw( Components::CCMException )\n" );
		out_.write( "{\n" );
		out_.write( "    return new CCM_" + impl + "();\n" );
		out_.write( "}\n" );
		out_.write( "\n" );

		out_.write( "extern \"C\"\n" );
		out_.write( "{\n" );
		out_.write( "    Components::HomeExecutorBase * create_" + DYNAMIC_ROUTING_NAME + "_home()\n" );
		out_.write( "    {\n" );
		out_.write( "        return new " + home_impl + "();\n" );
		out_.write( "    }\n" );
		out_.write( "}\n" );
		out_.write( "\n" );

		closeFile();
	}
	
	private void generateEnvironmentComponent( String resultsPath ) throws IOException {
		String componentDirectory = resultsPath + "/" + ENVIRONMENT_NAME;
		( new File( componentDirectory ) ).mkdir();
		
		String impl = ENVIRONMENT_NAME + "_impl";

		openFile( componentDirectory, impl, HPP_EXTENSION );
		
		writeCppHeaderPrelude( helper_.getApplication().getModule().getName() );
		
		out_.write( "#include \"microccm_" + ENVIRONMENT_NAME + "_executor." + HPP_EXTENSION + "\"\n" );
		out_.write( "#include \"microccm_" + ENVIRONMENT_NAME + "_CORBA." + HPP_EXTENSION + "\"\n" );
		out_.write( "\n" );
		
		out_.write( "class CCM_" + impl + "\n" );
		out_.write( "    : public CCM_" + ENVIRONMENT_NAME + "\n" );
		out_.write( "    , public CCM_" + ENVIRONMENT_NAME + "_SessionComponent_impl\n" );
		if( STATIC_DEPLOYMENT ) {
			out_.write( "    , public CCM_SimulationComponent\n" );
		}
		out_.write( "{\n" );
		out_.write( "public:\n" );
		out_.write( "    CCM_" + impl + "();\n" );
		out_.write( "\n" );
		out_.write( "    void configuration_complete()\n" );
		out_.write( "        throw( Components::InvalidConfiguration ) {}\n" );
		out_.write( "\n" );
		out_.write( "    void shutdown() {}\n" );
		out_.write( "\n" );

		if( STATIC_DEPLOYMENT ) {
			out_.write( "    void run();\n" );
			out_.write( "\n" );
		}

		HashSet< Port > ports = new HashSet< Port >();
		// We also provide the internal consumer ports (those used for dynamic
		// configuration) so that they can be examined
		//ports.addAll( helper_.getApplicationExternalConsumerPorts() );
		ports.addAll( helper_.getApplicationConsumerPorts() );
		ports.addAll( helper_.getApplicationExternalProducerPorts() );
		
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				// StreamType
				StreamType s = (StreamType) t;
				if( p instanceof ProducerPort ) {
					out_.write( "public:\n" );
					out_.write( "    void new_" + p.getName() + "_value( "
							  + getCorbaType( s.getType().getName() ) +" value );\n" );
					//out_.write( "    " + STI_MODULE_NAME + "::CCM_"
					out_.write( "    " + "CCM_"
							  + getStreamType( s.getType().getName() ) + "_ptr "
							  + "get_" + p.getName() + "() {\n" );
					out_.write( "        return &" + p.getName() + "_provider_;\n" );
					out_.write( "    }\n" );
					out_.write( "\n" );
					out_.write( "private:\n" );
					out_.write( "    struct Provides_" + p.getName() + "\n" );
					//out_.write( "        : " + STI_MODULE_NAME + "::CCM_"
					out_.write( "        : " + "CCM_"
							  + getStreamType( s.getType().getName() ) + "\n" );
					out_.write( "    {\n" );
					out_.write( "        Provides_" + p.getName()
							  + "( CCM_" + impl + " * c )\n" );
					out_.write( "            : component_( c ) {}\n" );
					out_.write( "\n" );
					out_.write( "        void put( "
							  + getCorbaType( s.getType().getName() )
							  + " value );\n" );
					out_.write( "\n" );
					out_.write( "        CCM_" + impl + " * component_;\n" );
					out_.write( "    };\n" );
					out_.write( "\n" );
					out_.write( "    Provides_" + p.getName() + " "
							  + p.getName() + "_provider_;\n" );
					out_.write( "\n" );
				}
			}
			else {
				// EventType
				if( p instanceof ConsumerPort ) {
					EventType e = (EventType) p.getTokenType();
					out_.write( "    // GENERATED FROM : consumes " + e.getName() + " "
							  + p.getName() + ";\n" );
					out_.write( "    void push_" + p.getName() + "( "
							  + e.getModule().getName() + "::"
							  + e.getName() + " * evt );\n" );
					out_.write( "\n" );
				}
			}
		}
		
		if( STATIC_DEPLOYMENT ) {
			out_.write( "public:\n" );
			out_.write( "        CCM_SimulationComponent_ptr get_start() {\n" );
			out_.write( "            return this;\n" );
			out_.write( "        }\n" );
			out_.write( "\n" );
		}

		out_.write( "\n" );
		out_.write( "};\n" );
		out_.write( "\n" );

		String home = ENVIRONMENT_NAME + "Home";
		String home_impl = home + "_impl";

		out_.write( "class " + home_impl + "\n" );
		out_.write( "    : public CCM_" + home + "\n" );
		out_.write( "{\n" );
		out_.write( "public:\n" );
		out_.write( "    Components::EnterpriseComponent_ptr create()\n" );
		out_.write( "        throw( Components::CCMException );\n" );
		out_.write( "};\n" );
		out_.write( "\n" );
		
		out_.write( "extern \"C\"\n" );
		out_.write( "{\n" );
		out_.write( "    Components::HomeExecutorBase * create_" + ENVIRONMENT_NAME + "_home();\n" );
		out_.write( "}\n" );
		out_.write( "\n" );

		writeCppHeaderPostlude();
		
		closeFile();
		
		openFile( componentDirectory, impl, CPP_EXTENSION );
		
		writeCppCopyrightComment();
		
		out_.write( "#include <string>\n" );
		out_.write( "#include <list>\n" );
		out_.write( "\n" );
		out_.write( "#include \"" + impl + "." + HPP_EXTENSION + "\"\n" );
		out_.write( "\n" );

		// Constructor

		out_.write( "CCM_" + impl + "::CCM_" + impl + "()\n" );
		boolean first_port = true;
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				if( p instanceof ProducerPort ) {
					if( first_port ) {
						out_.write( "    : " );
						first_port = false;
					}
					else {
						out_.write( "    , " );
					}
					out_.write( p.getName() + "_provider_( this )\n" );
				}
			}
		}
		out_.write( "{\n" );
		out_.write( "}\n" );
		out_.write( "\n" );
		
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				if( p instanceof ProducerPort ) {
					StreamType s = (StreamType) p.getTokenType();
					out_.write( "void CCM_" + impl + "::new_"
							  + p.getName() + "_value( "
							  + getCorbaType( s.getType().getName() )
							  + " value )\n" );
					out_.write( "{\n" );
					out_.write( "}\n" );
					out_.write( "\n" );
					
					out_.write( "void CCM_" + impl + "::Provides_" + p.getName()
							  + "::put( "
							  + getCorbaType( s.getType().getName() )
							  + " value )\n" );
					out_.write( "{\n" );
					out_.write( "    component_->new_" + p.getName() + "_value( value );\n" );
					out_.write( "}\n" );
					out_.write( "\n" );
				}
			}
			else {
				// EventType
				if( p instanceof ConsumerPort ) {
					EventType e = (EventType) p.getTokenType();
					out_.write( "void CCM_" + impl + "::push_" + p.getName() + "( "
							  + e.getModule().getName() + "::"
							  + e.getName() + " * evt )\n" );
					out_.write( "{\n" );
					out_.write( "}\n" );
					out_.write( "\n" );
				}
			}
		}
		out_.write( "\n" );
		
		if( STATIC_DEPLOYMENT ) {
			out_.write( "struct CCM_Object_instance {\n" );
			out_.write( "    std::string name_instance;\n" );
			out_.write( "    Components::CCMObject_ptr ccm_instance;\n" );
			out_.write( "};\n" );
			out_.write( "\n" );
			
			out_.write( "void client( std::list< CCM_Object_instance > & the_list )\n" );
			out_.write( "{\n" );
			out_.write( "    SimulationComponent_ptr start = 0;\n" );
			out_.write( "\n" );
			out_.write( "    for( std::list< CCM_Object_instance >::iterator c = the_list.begin();\n" );
			out_.write( "         c != the_list.end();\n" );
			out_.write( "         ++c ) {\n" );
			out_.write( "        if( c->name_instance == \"" + ENVIRONMENT_NAME + "\" ) {\n" );
			out_.write( "            CORBA::Object_ptr start_facet = c->ccm_instance->provide_facet( \"start\" );\n" );
			out_.write( "            start = SimulationComponent::_narrow( start_facet );\n" );
			out_.write( "        }\n" );
			out_.write( "    }\n" );
			out_.write( "\n" );
			out_.write( "    if( start != 0 ) {\n" );
			out_.write( "        start->run();\n" );
			out_.write( "    }\n" );
			out_.write( "}\n" );
			out_.write( "\n" );
	
			out_.write( "void CCM_" + impl + "::run()\n" );
			out_.write( "{\n" );
			out_.write( "}\n" );
			out_.write( "\n" );
		}
		
		out_.write( "//---------------------------------------------------------------------\n\n" );

		out_.write( "Components::EnterpriseComponent_ptr\n" );
		out_.write( home_impl + "::create() throw( Components::CCMException )\n" );
		out_.write( "{\n" );
		out_.write( "    return new CCM_" + impl + "();\n" );
		out_.write( "}\n" );
		out_.write( "\n" );

		out_.write( "extern \"C\"\n" );
		out_.write( "{\n" );
		out_.write( "    Components::HomeExecutorBase * create_" + ENVIRONMENT_NAME + "_home()\n" );
		out_.write( "    {\n" );
		out_.write( "        return new " + home_impl + "();\n" );
		out_.write( "    }\n" );
		out_.write( "}\n" );
		out_.write( "\n" );

		closeFile();
	}

	private EmfAdlvHelper helper_;
}
